
/*
 * Licensed Materials - Property of IBM
 *
 * trousers - An open source TCG Software Stack
 *
 * (C) Copyright International Business Machines Corp. 2004
 *
 */


#ifdef WIN32
#include <ws2tcpip.h>
#include <windows.h>
#endif
#include <stdlib.h>
#include <stdio.h>
#ifndef WIN32
#include <unistd.h>
#endif
#include <string.h>
#include <sys/types.h>
#ifndef WIN32
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netdb.h>
#include <pwd.h>
#endif
#if (defined (__OpenBSD__) || defined (__FreeBSD__))
#include <netinet/in.h>
#endif
#ifndef WIN32
#include <arpa/inet.h>
#endif
#include <errno.h>
#include <getopt.h>
#include "trousers/tss.h"
#include "trousers_types.h"
#include "tcs_tsp.h"
#include "tcs_utils.h"
#include "tcs_int_literals.h"
#include "capabilities.h"
#include "tcslog.h"
#include "tcsd_wrap.h"
#include "tcsps.h"
#include "tcsd.h"
#include "req_mgr.h"
#ifdef WIN32
#include "win32dep.h"
#include "libc_interface.h"
#define SERVICE_NAME		"tcsd"
#define WIN32_LEAN_AND_MEAN
#include <tchar.h>
#endif

struct tcsd_config tcsd_options;
struct tpm_properties tpm_metrics;
static volatile int hup = 0, term = 0;
extern char *optarg;
int sd = -1;

static void
tcsd_shutdown(void)
{
	/* order is important here:
	 * allow all threads to complete their current request */
	tcsd_threads_final();
	PS_close_disk_cache();
	auth_mgr_final();
	(void)req_mgr_final();
	conf_file_final(&tcsd_options);
	EVENT_LOG_final();
}

static void
tcsd_signal_term(int signal)
{
	term = 1;
	close(sd);
	sd = -1;
}

void
tcsd_signal_hup(int signal)
{
	hup = 1;
}

#ifdef WIN32
int serviceflag = 0;
SERVICE_STATUS_HANDLE serviceStatusHandle;
HANDLE killServiceEvent = INVALID_HANDLE_VALUE;
BOOL serviceRunning = 0;
BOOL servicePaused = 0;
DWORD serviceCurrentStatus;

/************************************************************************/
/* UpdateSCMStatus							*/
/************************************************************************/
BOOL UpdateSCMStatus (DWORD dwCurrentState,
					  DWORD dwWin32ExitCode,
					  DWORD dwServiceSpecificExitCode,
					  DWORD dwCheckPoint,
					  DWORD dwWaitHint)
{
	BOOL success;
	SERVICE_STATUS serviceStatus;

	if (serviceflag == 0)
		return TRUE;

	serviceCurrentStatus = dwCurrentState;
	serviceStatus.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
	serviceStatus.dwCurrentState = dwCurrentState;

	if (dwCurrentState == SERVICE_START_PENDING)
		serviceStatus.dwControlsAccepted = 0;
	else {
		serviceStatus.dwControlsAccepted =
			SERVICE_ACCEPT_STOP |
			//SERVICE_ACCEPT_PAUSE_CONTINUE |
			SERVICE_ACCEPT_SHUTDOWN;
	}

	if (dwServiceSpecificExitCode == 0)
		serviceStatus.dwWin32ExitCode = dwWin32ExitCode;
	else
		serviceStatus.dwWin32ExitCode = ERROR_SERVICE_SPECIFIC_ERROR;

	serviceStatus.dwServiceSpecificExitCode = dwServiceSpecificExitCode;
	serviceStatus.dwCheckPoint = dwCheckPoint;
	serviceStatus.dwWaitHint = dwWaitHint;

	success = SetServiceStatus (serviceStatusHandle, &serviceStatus);
	if (!success)
		return FALSE;

	return success;
}

BOOL signal_handler(DWORD event)
{
	tcsd_signal_term(0);
	return TRUE;
}

/************************************************************************/
/* ServiceCtrlHandler							*/
/************************************************************************/
VOID ServiceCtrlHandler (DWORD controlCode)
{
	BOOL success;
	switch(controlCode)
	{

	case SERVICE_CONTROL_PAUSE:
		break;
	case SERVICE_CONTROL_CONTINUE:
		break;
	case SERVICE_CONTROL_INTERROGATE:
		break;
	case SERVICE_CONTROL_STOP:
	case SERVICE_CONTROL_SHUTDOWN:
		success = UpdateSCMStatus(SERVICE_STOP_PENDING, NO_ERROR, 0, 1, 5000);
		signal_handler(CTRL_CLOSE_EVENT);
		return;
	default:
		break;
	}
	UpdateSCMStatus(serviceCurrentStatus, NO_ERROR, 0, 0, 0);
}
#endif

static TSS_RESULT
signals_init(void)
{
#ifndef WIN32
	int rc;
	sigset_t sigmask;
	struct sigaction sa;

	sigemptyset(&sigmask);
	if ((rc = sigaddset(&sigmask, SIGTERM))) {
		LogError("sigaddset: %s", strerror(errno));
		return TCSERR(TSS_E_INTERNAL_ERROR);
	}
	if ((rc = sigaddset(&sigmask, SIGHUP))) {
		LogError("sigaddset: %s", strerror(errno));
		return TCSERR(TSS_E_INTERNAL_ERROR);
	}

	if ((rc = THREAD_SET_SIGNAL_MASK(SIG_UNBLOCK, &sigmask, NULL))) {
		LogError("Setting thread signal mask: %s", strerror(rc));
		return TCSERR(TSS_E_INTERNAL_ERROR);
	}

	sa.sa_flags = 0;
	sigemptyset(&sa.sa_mask);
	sa.sa_handler = tcsd_signal_term;
	if ((rc = sigaction(SIGTERM, &sa, NULL))) {
		LogError("signal SIGTERM not registered: %s", strerror(errno));
		return TCSERR(TSS_E_INTERNAL_ERROR);
	}

	sa.sa_handler = tcsd_signal_hup;	
	if ((rc = sigaction(SIGHUP, &sa, NULL))) {
		LogError("signal SIGHUP not registered: %s", strerror(errno));
		return TCSERR(TSS_E_INTERNAL_ERROR);
	}
#else
	if (!SetConsoleCtrlHandler((PHANDLER_ROUTINE)signal_handler,TRUE)) {
		LogError("SetConsoleCtrlHandler() failed: %s", strerror(errno));
		return TCSERR(TSS_E_INTERNAL_ERROR);
	}
	return 0;
#endif

	return TSS_SUCCESS;
}

static TSS_RESULT
tcsd_startup(void)
{
	TSS_RESULT result;

#ifdef TSS_DEBUG
	/* Set stdout to be unbuffered to match stderr and interleave output correctly */
	setvbuf(stdout, (char *)NULL, _IONBF, 0);
#endif

	if ((result = signals_init()))
		return result;

	if ((result = conf_file_init(&tcsd_options)))
		return result;

	if ((result = tcsd_threads_init())) {
		conf_file_final(&tcsd_options);
		return result;
	}

	if ((result = req_mgr_init())) {
		conf_file_final(&tcsd_options);
		return result;
	}

	if ((result = ps_dirs_init())) {
		conf_file_final(&tcsd_options);
		(void)req_mgr_final();
		return result;
	}

	result = PS_init_disk_cache();
	if (result != TSS_SUCCESS) {
		conf_file_final(&tcsd_options);
		(void)req_mgr_final();
		return result;
	}

	if ((result = get_tpm_metrics(&tpm_metrics))) {
		conf_file_final(&tcsd_options);
		PS_close_disk_cache();
		(void)req_mgr_final();
		return result;
	}

	/* must happen after get_tpm_metrics() */
	if ((result = auth_mgr_init())) {
		conf_file_final(&tcsd_options);
		PS_close_disk_cache();
		(void)req_mgr_final();
		return result;
	}

	result = EVENT_LOG_init();
	if (result != TSS_SUCCESS) {
		auth_mgr_final();
		conf_file_final(&tcsd_options);
		PS_close_disk_cache();
		(void)req_mgr_final();
		return result;
	}

	result = owner_evict_init();
	if (result != TSS_SUCCESS) {
		auth_mgr_final();
		conf_file_final(&tcsd_options);
		PS_close_disk_cache();
		(void)req_mgr_final();
		return result;
	}

	return TSS_SUCCESS;
}


void
usage(void)
{
	fprintf(stderr, "\tusage: tcsd [-f] [-h]\n\n");
	fprintf(stderr, "\t-f|--foreground\trun in the foreground. Logging goes to stderr "
			"instead of syslog.\n");
	fprintf(stderr, "\t-e| attempts to connect to software TPMs over TCP");
	fprintf(stderr, "\t-h|--help\tdisplay this help message\n");
	fprintf(stderr, "\n");
}

static TSS_RESULT
reload_config(void)
{
	TSS_RESULT result;
	hup = 0;

	// FIXME: reload the config - work in progress
	result = TSS_SUCCESS;

	return result;
}


void
tcsd_main(int argc, char **argv)
{
	struct sockaddr_in serv_addr, client_addr;
	TSS_RESULT result = TSS_SUCCESS;
	int newsd, c;
	socklen_t client_len;
	char *hostname = NULL;
	struct hostent *client_hostent = NULL;
	
#ifdef WIN32
	BOOL success = 0;
	WSADATA wsaData;

	/* Registers ServiceCtrlHandler function to handle service control 
	 * requests. The return value is the service status handle. */
	if (serviceflag == 1) {
		serviceStatusHandle = RegisterServiceCtrlHandler((LPTSTR)SERVICE_NAME,
			(LPHANDLER_FUNCTION) ServiceCtrlHandler);
		if (serviceStatusHandle == 0) {
			LogError("RegisterServiceCtrlHandler() failed: %s", strerror(errno));
			goto outlast;
		}
	}

	/* SCM is notified that the service is in state START PENDING */
	success = UpdateSCMStatus(SERVICE_START_PENDING, NO_ERROR, 0, 1, 5000);
	if (!success) {
		LogError("UpdateSCMStatus() failed");
		goto outlast;
	}

	result = WSAStartup(MAKEWORD(2,2), &wsaData);
	if(result) {
		LogError("WSAStartup failed.");	
		goto outlast;
	}
#endif

	if ((result = tcsd_startup())) {
		LogError("tcsd_startup failed: %d", result);
		goto outlast;
	}

	sd = socket(AF_INET, SOCK_STREAM, 0);
	if (sd < 0) {
		LogError("Failed socket: %s", strerror(errno));
		goto out;
	}

	memset(&serv_addr, 0, sizeof (serv_addr));
	serv_addr.sin_family = AF_INET;
	serv_addr.sin_port = htons(tcsd_options.port);

	/* If no remote_ops are defined, restrict connections to localhost
	 * only at the socket. */
	if (tcsd_options.remote_ops[0] == 0)
		serv_addr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
	else
		serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);

	c = 1;
	setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, &c, sizeof(c));
	if (bind(sd, (struct sockaddr *) &serv_addr, sizeof (serv_addr)) < 0) {
		LogError("Failed bind: %s", strerror(errno));
		goto out;
	}

	if (listen(sd, TCSD_MAX_SOCKETS_QUEUED) < 0) {
		LogError("Failed listen: %s", strerror(errno));
		goto out;
	}
	client_len = (unsigned)sizeof(client_addr);

#ifdef WIN32
	/* SCM is notified that the service is still in the state RUNNING */
	success = UpdateSCMStatus(SERVICE_RUNNING, NO_ERROR, 0, 0, 0);
	if (!success){
		LogError("UpdateSCMStatus() failed");
		goto out;
	}
#endif
	LogInfo("%s: TCSD up and running.", PACKAGE_STRING);
	do {
		newsd = accept(sd, (struct sockaddr *) &client_addr, &client_len);
		if (newsd < 0) {
			if (errno == EINTR) {
				if (term)
					break;
				else if (hup) {
					if (reload_config() != TSS_SUCCESS)
						LogError("Failed reloading config");
				}
				continue;
			} else {
				LogError("Failed accept: %s", strerror(errno));
				continue;
			}
		}
		LogDebug("accepted socket %i", newsd);

		if ((client_hostent = gethostbyaddr((char *) &client_addr.sin_addr,
						    sizeof(client_addr.sin_addr),
						    AF_INET)) == NULL) {
			char buf[16];
                        UINT32 addr = htonl(client_addr.sin_addr.s_addr);

                        snprintf(buf, 16, "%d.%d.%d.%d", (addr & 0xff000000) >> 24,
                                 (addr & 0x00ff0000) >> 16, (addr & 0x0000ff00) >> 8,
                                 addr & 0x000000ff);

			LogWarn("Host name for connecting IP %s could not be resolved", buf);
			hostname = strdup(buf);
		} else {
			hostname = strdup(client_hostent->h_name);
		}

		tcsd_thread_create(newsd, hostname);
		hostname = NULL;
		if (hup) {
			if (reload_config() != TSS_SUCCESS)
				LogError("Failed reloading config");
		}
	} while (term ==0);

	/* To close correctly, we must receive a SIGTERM */
out:
	if (term != 1)
		result = GetLastError();
	if (sd != -1)
		close(sd);
	tcsd_shutdown();
#ifdef WIN32
	WSACleanup();
outlast:
	UpdateSCMStatus(SERVICE_STOPPED, result,0,0,0);
#endif
	return;
} 

int
main(int argc, char **argv)
{
	int c, option_index = 0;
#if (!defined SOLARIS && !defined WIN32)
	struct passwd *pwd;
#endif

	struct option long_options[] = {
		{"help", 0, NULL, 'h'},
		{"foreground", 0, NULL, 'f'},
		{0, 0, 0, 0}
	};

	unsetenv("TCSD_USE_TCP_DEVICE");
	while ((c = getopt_long(argc, argv, "fhe", long_options, &option_index)) != -1) {
		switch (c) {
			case 'f':
				setenv("TCSD_FOREGROUND", "1", 1);
				break;
			case 'h':
			case 'e':
				setenv("TCSD_USE_TCP_DEVICE", "1", 1);
				break;
			default:
				usage();
				return -1;
				break;
		}
	}

#if !defined SOLARIS && !defined WIN32
	pwd = getpwnam(TSS_USER_NAME);
	if (pwd == NULL) {
		if (errno == 0) {
			LogError("User \"%s\" not found, please add this user"
					" manually.", TSS_USER_NAME);
		} else {
			LogError("getpwnam(%s): %s", TSS_USER_NAME, strerror(errno));
		}
		return TCSERR(TSS_E_INTERNAL_ERROR);
	}
	setuid(pwd->pw_uid);
#endif
	if (getenv("TCSD_FOREGROUND") == NULL) {
#ifdef WIN32
		SERVICE_TABLE_ENTRY serviceTable[] = {
			{ (LPTSTR)SERVICE_NAME, (LPSERVICE_MAIN_FUNCTION)tcsd_main },
			{ NULL, NULL } };
		int result;

		serviceflag = 1;

		/* Connects the main thread of a service process to the service 
		* control manager, which causes the thread to be the service control 
		* dispatcher thread for the calling process (in our case the ServiceMain) 
		* If the function succeeds, the return value is nonzero */
		result = StartServiceCtrlDispatcher(serviceTable);
		if (result == 0) {
	  		LogError("StartServiceCtrlDispatcher() failed: %s", strerror(errno));
			return -1;
		}

		return TSS_SUCCESS;
#else
		if (daemon(0, 0) == -1) {
			perror("daemon");
			return -1;
		}
#endif
	}
	tcsd_main(0, NULL);
	return 0;
}
